'use strict'

const Redis = require('redis')
const util = require('util')
const Pool = require('generic-pool')

class RedisPool {
    constructor(clientOptions, poolOptions) {
        const getClientOptions = function () {
            return clientOptions
        }

        this._getClientOptions = getClientOptions

        const factory = {
            async create() {
                const client = Redis.createClient(clientOptions)
                await client.pingAsync()
                return client
            },

            /**
             * 
             * @param {Redis.RedisClient} client 
             */
            destroy(client) {
                return client.quitAsync()
            }
        }

        this.pool = Pool.createPool(factory, poolOptions)
    }

    async multi(scope) {
        const client = Redis.createClient(this._getClientOptions())
        const multi = client.multi()
        let ret, error
        try {
            ret = await scope(multi)
        }
        catch (err) {
            error = err
        }
        finally {
            await client.quitAsync()
        }

        if (error) {
            throw error
        }

        return ret
    }

    batch(scope) {
        return this.multi(scope)
    }
}

const poolMethods = Object.keys(Redis.RedisClient.prototype)

poolMethods.forEach(name => {
    if (typeof (Redis.RedisClient.prototype[name]) !== 'function') {
        return
    }
    if (name.endsWith('Async')) {
        return
    }
    RedisPool.prototype[name] = function (...args) {
        let callback
        if (args.length > 0) {
            callback = args[args.length - 1]
            if (typeof callback === 'function') {
                args.pop()
            }
        }

        this.pool.acquire().then(client => {
            client[name].call(client, ...args, (...rets) => {
                this.pool.release(client).then(() => {
                    if (typeof callback === 'function') {
                        callback(...rets)
                    }
                })
            })
        }).catch(err => {
            if (typeof callback === 'function') {
                callback(err)
            }
        })
    }

    RedisPool.prototype[`${name}Async`] = async function (...args) {
        let client = await this.pool.acquire()
        let ret, error
        try {
            ret = await client[`${name}Async`].apply(client, args)
        }
        catch (err) {
            error = err
        }
        finally {
            await this.pool.release(client)
        }
        if (error) {
            throw error
        }
        
        return ret
    }
})

module.exports = RedisPool